import torch
import torch.nn as nn
import torch.nn.functional as F
import dgl
from dgl import function as fn
from . import BaseModel,register_model


@register_model('KGAT')
class KGAT(BaseModel):
    r"""
    This model KGAT was introduced in `KGAT <https://arxiv.org/pdf/1905.07854v2.pdf>`__.

    It involves some careful design.

    Embedding Layer:

    Knowledge graph embedding is an effective way to parameterize entities and relations as vector representations.
    KGAT employ TransR, a widely used method, on CKG.

    Attentive Embedding Propagation Layers:

    A single attentive embedding propagation layer consists of three components: information propagation, knowledge-aware attention, and information aggregation.

    1. Information Propagation:

    Considering an entity :math:`h`, we use :math:`\mathcal{N}_{h} = {(h,r,t)|(h,r,t)\in\mathcal{G}}` to denote the set of triplets where :math:`h` is the head entity, termed ego-network.
    To characterize the first-order connectivity structure of entity :math:`h`, we compute the linear combination of :math:`h`’s ego-network:

    :math:`e_{\mathcal{N}_{h}}=\sum_{(h,r,t)\in\mathcal{N}_{h}}\pi(h,r,t)e_{t}`

    where :math:`\pi(h,r,t)` controls the decay factor on each propagation on edge :math:`(h,r,t)`, indicating how much information being propagated from :math:`t` to :math:`h` conditioned to relation :math:`r`.

    2. Knowledge-aware Attention:

    GAT implement :math:`\pi(h,r,t)` via relational attention mechanism, which is formulated as follows:

    :math:`\pi(h,r,t)=(\mathcal{N}_{r}\mathcal{e}_{t})^{\mathsf{T}}tanh((\mathcal{W}_{r}\mathcal{e}_{h}+e_{r}))`

    This makes the attention score dependent on the distance between :math:`e_h` and :math:`e_t` in the relation :math:`r`’s space
    Hereafter, GAT normalize the coefficients across all triplets connected with :math:`h` by adopting the softmax function:

    :math:`\pi(h,r,t)=\frac{exp(\pi(h,r,t))}{\sum_{(h,r',t')\in\mathcal{N}_{h}}exp(\pi(h,r',t'))}`

    As a result, the final attention score is capable of suggesting which neighbor nodes should be given more attention to capture collaborative signals.

    3. Information Aggregation:
    The final phase is to aggregate the entity representation :math:`e_h` and its ego-network representations :math:`e_{\mathcal{N}_h}` as the new representation of entity :math:`h`

    GCN Aggregator : :math:`\mathcal{f}_{GCN}=LeakyReLU(\mathcal{W}(e_h+e_{\mathcal{N}_h}))`

    GraphSage Aggregator : :math:`\mathcal{f}_{GraphSage}=LeakyReLU(\mathcal{W}(e_h||e_{\mathcal{N}_h}))`

    Bi-Interaction Aggregator : :math:`\mathcal{f}_{Bi-Interaction}=LeakyReLU(\mathcal{W}(e_h+e_{\mathcal{N}_h})+LeakyReLU(\mathcal{W}(e_h\odote_{\mathcal{N}_h})`

    High-order Propagation:
    We can further stack more propagation layers to explore the high-order connectivity information, gathering the information propagated from the higher-hop neighbors.

    :math:`e_{h}^{(l)}=\mathcal{f}(e_{h}^{(l-1)}_{\mathcal{N}_h})`

    Model Prediction：

    After performing :math:`L` layers, we obtain multiple representations for user node :math:`u`, namely :math:`{e_{u}^{(1)},...,{e_{u}^{(L)}}`; Analogous to item node i, :math:`{e_{i}^{(1)},...,{e_{i}^{(L)}}` are obtained.
    GAT hence adopt the layer-aggregation mechanism to concatenate the representations at each step into a single vector, as follows:

    :math:`e^*_u=e_u^{(0)}||...||e_u^{(L)},e^*_i=e_i^{(0)}||...||e_i^{(L)}`

    Finally, we conduct inner product of user and item representations, so as to predict their matching score:

    :math:`\check{\mathcal{y}}(u,i)=e^*_u`\mathsf{T}e^*_i`

    Parameters
    ----------
    entity_dim ：User / entity Embedding size
    relation_dim ： Relation Embedding size
    aggregation_type ： Specify the type of the aggregation layer from {gcn, graphsage, bi-interaction}
    conv_dim_list ： Output sizes of every aggregation layer
    mess_dropout ： Dropout probability w.r.t. message dropout for each deep layer. 0: no dropout
    """
    @classmethod
    def build_model_from_args(cls,args,hg):
        return cls(args=args)

    def __init__(self,args):
        super(KGAT, self).__init__()
        if args.aggregation_type not in ['bi-interaction','gcn','graphsage']:
            raise KeyError('Aggregator type {} not supported.'.format(args.aggregation_type))
        self.use_pretrain = args.use_pretrain
        self.entity_dim = args.entity_dim
        self.relation_dim = args.relation_dim
        self.aggregation_type = args.aggregation_type
        self.conv_dim_list = [args.entity_dim] + eval(args.conv_dim_list)
        self.mess_dropout = eval(args.mess_dropout)
        self.n_layers = len(eval(args.conv_dim_list))

        self.kg_l2loss_lambda = args.kg_l2loss_lambda
        self.cf_l2loss_lambda = args.cf_l2loss_lambda

        self.aggregator_layers = nn.ModuleList()
        for k in range(self.n_layers):
            self.aggregator_layers.append(KGAT_Aggregator(self.conv_dim_list[k], self.conv_dim_list[k + 1], self.mess_dropout[k], self.aggregation_type))

    def set_parameters(self, n_users, n_entities, n_relations, user_pre_embed=None, item_pre_embed=None):

        self.n_users = n_users
        self.n_entities = n_entities
        self.n_relations = n_relations

        # Embedding
        self.relation_embed = nn.Embedding(self.n_relations, self.relation_dim)
        self.entity_user_embed = nn.Embedding(self.n_entities + self.n_users, self.entity_dim)
        if (self.use_pretrain == 1) and (user_pre_embed is not None) and (item_pre_embed is not None):
            other_entity_embed = nn.Parameter(torch.Tensor(self.n_entities - item_pre_embed.shape[0], self.entity_dim))
            nn.init.xavier_uniform_(other_entity_embed, gain=nn.init.calculate_gain('relu'))
            entity_user_embed = torch.cat([item_pre_embed, other_entity_embed, user_pre_embed], dim=0)
            self.entity_user_embed.weight = nn.Parameter(entity_user_embed)

        self.W_R = nn.Parameter(torch.Tensor(self.n_relations, self.entity_dim, self.relation_dim))
        nn.init.xavier_uniform_(self.W_R, gain=nn.init.calculate_gain('relu'))

    '''
    # DGL: dgl-cu101(0.4.3)
    # We will get different results when using the function `fn.sum`, and the randomness is due to `atomicAdd`.
    # Use custom function to ensure deterministic behavior.
    '''

    def edge_softmax_fix(self,graph,score):

        def reduce_sum(nodes):
            accum = torch.sum(nodes.mailbox['temp'], 1)
            return {'out_sum': accum}

        graph = graph.local_var()
        graph.edata['out'] = score
        graph.edata['out'] = torch.exp(graph.edata['out'])
        graph.update_all(fn.copy_e('out', 'temp'), reduce_sum)
        graph.apply_edges(fn.e_div_v('out', 'out_sum', 'out'))
        out = graph.edata['out']
        return out

    def att_score(self, edges):
        # Equation (4)
        r_mul_t = torch.matmul(self.entity_user_embed(edges.src['id']), self.W_r)                       # (n_edge, relation_dim)
        r_mul_h = torch.matmul(self.entity_user_embed(edges.dst['id']), self.W_r)                       # (n_edge, relation_dim)
        r_embed = self.relation_embed(edges.data['type'])                                               # (1, relation_dim)
        att = torch.bmm(r_mul_t.unsqueeze(1), torch.tanh(r_mul_h + r_embed).unsqueeze(2)).squeeze(-1)   # (n_edge, 1)
        return {'att': att}

    def compute_attention(self, g):
        g = g.local_var()
        for i in range(self.n_relations):
            edge_idxs = g.filter_edges(lambda edge: edge.data['type'] == i)
            self.W_r = self.W_R[i]  # [entity_dim, relation_dim]
            g.apply_edges(self.att_score, edge_idxs)

        # Equation (5)
        g.edata['att'] = self.edge_softmax_fix(g, g.edata.pop('att'))
        return g.edata.pop('att')

    def _L2_loss_mean(self,x):
        return torch.mean(torch.sum(torch.pow(x, 2), dim=1, keepdim=False) / 2.)

    def calc_kg_loss(self, h, r, pos_t, neg_t):   #f
        """
        h:      (kg_batch_size)
        r:      (kg_batch_size)
        pos_t:  (kg_batch_size)
        neg_t:  (kg_batch_size)
        """
        r_embed = self.relation_embed(r)                 # (kg_batch_size, relation_dim)
        W_r = self.W_R[r]                                # (kg_batch_size, entity_dim, relation_dim)

        h_embed = self.entity_user_embed(h)              # (kg_batch_size, entity_dim)
        pos_t_embed = self.entity_user_embed(pos_t)      # (kg_batch_size, entity_dim)
        neg_t_embed = self.entity_user_embed(neg_t)      # (kg_batch_size, entity_dim)

        r_mul_h = torch.bmm(h_embed.unsqueeze(1), W_r).squeeze(1)             # (kg_batch_size, relation_dim)
        r_mul_pos_t = torch.bmm(pos_t_embed.unsqueeze(1), W_r).squeeze(1)     # (kg_batch_size, relation_dim)
        r_mul_neg_t = torch.bmm(neg_t_embed.unsqueeze(1), W_r).squeeze(1)     # (kg_batch_size, relation_dim)

        # Equation (1)
        pos_score = torch.sum(torch.pow(r_mul_h + r_embed - r_mul_pos_t, 2), dim=1)     # (kg_batch_size)
        neg_score = torch.sum(torch.pow(r_mul_h + r_embed - r_mul_neg_t, 2), dim=1)     # (kg_batch_size)

        # Equation (2)
        kg_loss = (-1.0) * F.logsigmoid(neg_score - pos_score)
        kg_loss = torch.mean(kg_loss)

        l2_loss = self._L2_loss_mean(r_mul_h) + self._L2_loss_mean(r_embed) + self._L2_loss_mean(r_mul_pos_t) + self._L2_loss_mean(r_mul_neg_t)
        loss = kg_loss + self.kg_l2loss_lambda * l2_loss
        return loss

    def cf_embedding(self, mode, g):
        g = g.local_var()
        ego_embed = self.entity_user_embed(g.ndata['id'])
        all_embed = [ego_embed]

        for i, layer in enumerate(self.aggregator_layers):
            ego_embed = layer(mode, g, ego_embed)
            norm_embed = F.normalize(ego_embed, p=2, dim=1)
            all_embed.append(norm_embed)

        # Equation (11)
        all_embed = torch.cat(all_embed, dim=1)         # (n_users + n_entities, cf_concat_dim)
        return all_embed

    def cf_score(self, mode, g, user_ids, item_ids):
        """
        user_ids:   number of users to evaluate   (n_eval_users)
        item_ids:   number of items to evaluate   (n_eval_items)
        """
        all_embed = self.cf_embedding(mode, g)          # (n_users + n_entities, cf_concat_dim)
        user_embed = all_embed[user_ids]                # (n_eval_users, cf_concat_dim)
        item_embed = all_embed[item_ids]                # (n_eval_items, cf_concat_dim)

        # Equation (12)
        cf_score = torch.matmul(user_embed, item_embed.transpose(0, 1))    # (n_eval_users, n_eval_items)
        return cf_score

    def calc_cf_loss(self, mode, g, user_ids, item_pos_ids, item_neg_ids):
        """
        user_ids:       (cf_batch_size)
        item_pos_ids:   (cf_batch_size)
        item_neg_ids:   (cf_batch_size)
        """
        all_embed = self.cf_embedding(mode, g)                      # (n_users + n_entities, cf_concat_dim)
        user_embed = all_embed[user_ids]                            # (cf_batch_size, cf_concat_dim)
        item_pos_embed = all_embed[item_pos_ids]                    # (cf_batch_size, cf_concat_dim)
        item_neg_embed = all_embed[item_neg_ids]                    # (cf_batch_size, cf_concat_dim)

        # Equation (12)
        pos_score = torch.sum(user_embed * item_pos_embed, dim=1)   # (cf_batch_size)
        neg_score = torch.sum(user_embed * item_neg_embed, dim=1)   # (cf_batch_size)

        # Equation (13)
        cf_loss = (-1.0) * F.logsigmoid(pos_score - neg_score)
        cf_loss = torch.mean(cf_loss)

        l2_loss = self._L2_loss_mean(user_embed) + self._L2_loss_mean(item_pos_embed) + self._L2_loss_mean(item_neg_embed)
        loss = cf_loss + self.cf_l2loss_lambda * l2_loss
        return loss

    def forward(self, mode, *input):
        if mode == 'calc_att':
            return self.compute_attention(*input)
        if mode == 'calc_cf_loss':
            return self.calc_cf_loss(mode, *input)
        if mode == 'calc_kg_loss':
            return self.calc_kg_loss(*input)
        if mode == 'predict':
            return self.cf_score(mode, *input)


class KGAT_Aggregator(nn.Module):

    def __init__(self, in_dim, out_dim, dropout, aggregator_type):
        super(KGAT_Aggregator, self).__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.dropout = dropout
        self.aggregator_type = aggregator_type

        self.message_dropout = nn.Dropout(dropout)

        if aggregator_type == 'gcn':
            self.W = nn.Linear(self.in_dim, self.out_dim)       # W in Equation (6)
        elif aggregator_type == 'graphsage':
            self.W = nn.Linear(self.in_dim * 2, self.out_dim)   # W in Equation (7)
        elif aggregator_type == 'bi-interaction':
            self.W1 = nn.Linear(self.in_dim, self.out_dim)      # W1 in Equation (8)
            self.W2 = nn.Linear(self.in_dim, self.out_dim)      # W2 in Equation (8)
        else:
            raise NotImplementedError

        self.activation = nn.LeakyReLU()

    def forward(self, mode, g, entity_embed):
        g = g.local_var()
        g.ndata['node'] = entity_embed

        # Equation (3) & (10)
        # DGL: dgl-cu10.1(0.5.3)
        # Get different results when using `dgl.function.sum`, and the randomness is due to `atomicAdd`
        # Use `dgl.function.sum` when training model to speed up
        # Use custom function to ensure deterministic behavior when predicting
        if mode == 'predict':
            g.update_all(dgl.function.u_mul_e('node', 'att', 'side'), lambda nodes: {'N_h': torch.sum(nodes.mailbox['side'], 1)})
        else:
            g.update_all(dgl.function.u_mul_e('node', 'att', 'side'), dgl.function.sum('side', 'N_h'))

        if self.aggregator_type == 'gcn':
            # Equation (6) & (9)
            out = self.activation(self.W(g.ndata['node'] + g.ndata['N_h']))                    # (n_users + n_entities, out_dim)

        elif self.aggregator_type == 'graphsage':
            # Equation (7) & (9)
            out = self.activation(self.W(torch.cat([g.ndata['node'], g.ndata['N_h']], dim=1))) # (n_users + n_entities, out_dim)

        elif self.aggregator_type == 'bi-interaction':
            # Equation (8) & (9)
            out1 = self.activation(self.W1(g.ndata['node'] + g.ndata['N_h']))                  # (n_users + n_entities, out_dim)
            out2 = self.activation(self.W2(g.ndata['node'] * g.ndata['N_h']))                  # (n_users + n_entities, out_dim)
            out = out1 + out2
        else:
            raise NotImplementedError

        out = self.message_dropout(out)
        return out


import torch
import torch.nn as nn
from torch.nn import init
import torch.nn.functional as F
import numpy as np
from . import BaseModel, register_model


@register_model("Mg2vec")
class Mg2vec(BaseModel):
    r"""
    This is a model mg2vec from `mg2vec: Learning Relationship-Preserving Heterogeneous Graph Representations via
    Metagraph Embedding<https://ieeexplore.ieee.org/document/9089251>`__

    It contains following parts:

    Achieve the metagraph and metagraph instances by mining the raw graph. Please go to
    `DataMaker-For-Mg2vec<https://github.com/null-xyj/DataMaker-For-Mg2vec>`__ for more details.

    Initialize the embedding for every node and metagraph and adopt an unsupervised method to train the node embeddings
    and metagraph embeddings. In detail, for every node, we keep its embedding close to the metagraph it belongs to and
    far away from the metagraph we get by negative sampling.

    Every node and meta-graph can be represented as an n-dim vector.We define the first-order loss and second-order
    loss.
    First-Order Loss is for single core node in every meta-graph. We compute the dot product of the node embedding and
    the positive meta-graph embedding as the true logit. Then We compute the dot product of the node embedding and
    the sampled negative meta-graph embedding as the neg logit. We use the binary_cross_entropy_with_logits function to
    compute the first-order loss.
    Second-Order Loss consider two core nodes in every meta-graph. First, we cancat the two node's embedding, what is a
    2n-dim vector. Then we use a 2n*n matrix and an n-dim vector to map the 2n-dim vector to an n-dim vector. The map
    function is showed below:
    .. math::
        f(u,v) = RELU([u||v]W + b)
    u and v means the origin embedding of the two nodes, || is the concatenation operator. W is the 2n*n matrix and b is
    the n-dim vector. RELU is the an activation function. f(u,v) means the n-dim vector after transforming.
    Then, the computation of second-order loss is the same as the first-order loss.
    Finally, we use a parameter alpha to balance the first-order loss and second-order loss.
    .. math::
        L=(1-alpha)*L_1 + alpha*L_2

    After we train the node embeddings, we use the embeddings to complete the relation prediction task.
    The relation prediction task is achieved by edge classification task. If two nodes are connected with a relation, we
    see the relation as an edge. Then we can adopt the edge classification to complete relation prediction task.

    Parameters
    ----------
    node_num: int
        the number of core-nodes
    mg_num: int
        the number of meta-graphs
    emb_dimension: int
        the embedding dimension of nodes and meta-graphs
    unigram: float
        the frequency of every meta-graph, for negative sampling
    sample_num: int
        the number of sampled negative meta-graph

    """
    @classmethod
    def build_model_from_args(cls, args, hg):
        return cls(args.node_num, args.mg_num, args.emb_dimension, args.unigram, args.sample_num)

    def __init__(self, node_num, mg_num, emb_dimension, unigram, sample_num):
        super(Mg2vec, self).__init__()
        self.node_num = node_num
        self.mg_num = mg_num
        self.mg_unigrams = torch.tensor(unigram, dtype=torch.float64)
        self.sample_num = sample_num
        self.emb_dimension = emb_dimension
        self.n_embedding = nn.Embedding(node_num, emb_dimension, sparse=False)
        self.m_embedding = nn.Embedding(mg_num, emb_dimension, sparse=False)
        self.n_w_t = nn.Parameter(torch.empty([emb_dimension * 2, emb_dimension]), requires_grad=True)
        self.n_b = nn.Parameter(torch.empty(emb_dimension), requires_grad=True)

        init.xavier_normal_(self.n_embedding.weight.data)
        init.xavier_normal_(self.m_embedding.weight.data)
        init.xavier_normal_(self.n_w_t)
        init.constant_(self.n_b, 0)

    def forward(self, train_a, train_b, train_labels, train_freq, train_weight, device):
        batch_size = len(train_a)
        n_embed_a = self.n_embedding(train_a)
        n_embed_b = self.n_embedding(train_b)
        n_embed_con = torch.cat([n_embed_a, n_embed_b], dim=1)
        mask_o1 = torch.eq(train_a, train_b).type(torch.FloatTensor).reshape(batch_size, 1).to(device)
        mask_o2 = torch.not_equal(train_a, train_b).type(torch.FloatTensor).reshape(batch_size, 1).to(device)
        m_embed_pos = self.m_embedding(train_labels)
        neg_sample_id = torch.multinomial(self.mg_unigrams, min(self.mg_num, self.sample_num)).to(device)
        neg_m_embed = self.m_embedding(neg_sample_id)
        n_embed_o1 = n_embed_a * mask_o1
        n_embed_o2 = F.relu(torch.mm(n_embed_con, self.n_w_t) + self.n_b) * mask_o2
        n_embed = torch.add(n_embed_o1, n_embed_o2)
        true_logit = torch.sum((n_embed * m_embed_pos), dim=1, keepdim=True)
        neg_logit = torch.mm(n_embed, neg_m_embed.T)
        logit = torch.cat([true_logit, neg_logit], dim=1)
        labels = torch.cat([torch.ones_like(true_logit), torch.zeros_like(neg_logit)], dim=1)
        xent = torch.sum(F.binary_cross_entropy_with_logits(logit, labels, reduction='none'), dim=1, keepdim=True)
        unsupervised_loss = torch.mean(train_weight * (train_freq * xent))
        return unsupervised_loss

    def normalize_embedding(self):
        norm = torch.sqrt_(torch.sum(torch.square(self.n_embedding.weight.data), dim=1, keepdim=True))
        self.n_embedding.weight.data = self.n_embedding.weight.data / norm

        m_norm = torch.sqrt_(torch.sum(torch.square(self.m_embedding.weight.data), dim=1, keepdim=True))
        self.m_embedding.weight.data = self.m_embedding.weight.data / m_norm

    def save_embedding(self, id2node, file_name):
        self.normalize_embedding()
        embedding = self.n_embedding.weight.cpu().data.numpy()
        with open(file_name, 'w') as f:
            for nId, node in id2node.items():
                to_write = str(node) + ' ' + ' '.join(map(lambda x: str(x), embedding[nId])) + '\n'
                f.write(to_write)

    def save_embedding_np(self, file_name):
        self.normalize_embedding()
        embedding = self.n_embedding.weight.cpu().data.numpy()
        np.save(file_name, embedding)
